% ABM: LAILA Metainterpreter for ALIAS AGENTS: si occupa della comunicazione tra agenti.

% Ogni ALIAS AGENTS ha una parte come questa che accede alla blackboard dell'IP e si occupa
% della comunicazione tra agenti tramite linguaggio lAILA:  l'ABM che si mette in ascolto
% con una in(laila_query(WhoAmI,LailaQuery,AbdIn)) di eventuali richieste di altri ALIAS AGENT
% nella forma out(laila_query(WhoAmI,LailaQuery,AbdIn)).
% ES: se l'agente 'pippo' sull'IP1 vuole chiedere all'agente 'pluto' sull'IP2 di risolvere
% un goal G 'pippo' dovr effettuare una laila query di tipo comunicativo ( 'pluto' > G ).
% Nella knowledge base di 'pippo' ci sar l'informazione sull' IP in cui si trova 'pluto'
% quindi  l'ABM di 'pippo' far una out() remota che verr ricevuta dall'ABM di 'pluto' del tipo
% out(laila_query('pluto', G, AbdIn)).

% Se si vogliono usare pi agenti su di uno stesso IP, dal momento che non posso far 
% partire pi run_server sulla stessa porta devo usare oltre agli indirizzi IP anche
% le potre. Quindi fare set_this_port() prima di run_server e per identificare un'agente
% usare agent(Agent, Host, Port).

% LAILA server: fetches laila queries from the blackboard
% queries are in the form of tuples: laila_query(<MyName>,Query,InputDelta)
% The agent has got one blackboard (1 agent per IP in the system); the
% agent is the blackboard server, it fetches queries by means of 'in'
% primitives and solves them, putting an answer on the blackboard.
% answers to queries are tuples of the form: answer(<MyName>,Query,InputDelta,OutputDelta).
% for instance: laila_query('torroni',s1,[]), answer('torroni',s1,[],[m1])
% I assume that OutputDelta is funcition of (agent,query,inputDelta).

start:- bg(run_laila_server).

run_laila_server:-
	bg(server), % my blackboard is your blackboard
	iAm(WhoAmI), % my name:  nella kb dell'agente
	agent(WhoAmI,Host,Port), %  nella kb dell'agente
	set_that_port(Port),
 	there, % in the blackboard I share
	out(cont(0)), %  l'identificatore iniziale per le query.
	repeat, % loop...
	bg(senza_label),
	in(laila_query(Q,WhoAmI,LailaQuery,AbdIn)), % fetch (blocking)
	nl,
	println('Fetched a query for me. Issuing background execution...'), nl,
	bg(solve_laila_query(Q,WhoAmI,LailaQuery,AbdIn)), % background execution (non-blocking)
	fail. % ...forever

% Risolve query che non siano comunicazione, collaborazione o competizione
solve_laila_query(Q,WhoAmI,LailaQuery,AbdIn):-
	not(eq(Q,[comm,_,_])), % controllo che non sia una comunicazione,
	println(solve_laila_query_normale), nl,
	new_engine(laila(LailaQuery,AbdIn,AbdOut),AbdOut,Engine), % issue a thread (the thread then begins to solve the query)
	println('new_engine'), nl,
	ask_engine(Engine,Ans), % ask the thread for one answer   
	if(
         % se la risposta  del tipo the(X)
	   eq(Ans,the(X)),
         % allora...
	   eq(Answer,X),
         % altrimenti la risp  NO
	   eq(Answer,Ans)
	  ), 
	println('ask_engine:'),
	println(Answer), nl,
	out(answer(WhoAmI,LailaQuery,AbdIn,Answer)), % once you get it, put the answer on the blackboard
	
	% here the agent could be asked more solutions...
	
	stop_engine(Engine). % I'm done with that thread

% Risolve query di comunicazione
solve_laila_query(Q,WhoAmI,LailaQuery,AbdIn):-
	eq(Q,[comm,Agent,_]), % controllo che sia una comunicazione
	println(solve_laila_query_comunicazione), nl,
	new_engine(laila(LailaQuery,AbdIn,AbdOut),AbdOut,Engine), % issue a thread (the thread then begins to solve the query)
	println('new_engine'), nl,
	agent(Agent, Host, Port), %  nella kb dell'agente
	% set_host(Host),  da prob in locale
	set_that_port(Port),
	there,
	solvecom(Q,Engine).

solvecom(Q,Engine):-
	ask_engine(Engine,Ans), % ask the thread for one answer
	println('ask_engine:'),
	println(Ans), nl,
	out(ris(Q,Ans)),   
	testend(Q,Engine,Ans).

testend(Q,Engine,no):-
	   % potrei mettere or(eq(Ans,no),rd(halt)) per gestire l'halt
	   stop_engine(Engine).

testend(Q,Engine,Ans):-
	not(eq(Ans,no)),
	solvecom(Q,Engine).


% LAILA metainterpreter: queries are in the form: laila(LailaQuery,AbdIn,AbdOut)

% termination
laila([], AbdIn, AbdIn):-
	println('laila: terminated, returning '),
	println(AbdIn), nl.

% unroll goal
laila([A|B], AbdIn, AbdOut):-
	println('laila: unrolling goal'), nl,
	laila(A, AbdIn, AbdTemp),
	laila(B, AbdTemp, AbdOut).

% normal deduction
laila(Atom, AbdIn, AbdOut):-
	println('laila: performing normal deduction'), nl,
	abm(Atom, Body), % le clausole LAILA sono di questo tipo. 
	laila(Body, AbdIn, AbdOut).

% down-reflection
laila(down(Goal), AbdIn, AbdOut):-
	println('laila: solving down reflection...'), nl,
	solve_ARM_goal(Goal, AbdIn, AbdOut),
	% verifico consistenza con la query
	if(
	   is_list(Goal),
	   no_conflicts(Goal,AbdOut),
	   no_conflicts([Goal],AbdOut) 
	  ).

% communication
laila(ask(Agent, Goal), AbdIn, AbdOut):-
	println('laila: solving commuication...'), nl,
	label_comm(Q), 
	agent(Agent, Host, Port), %  nella kb dell'agente
	% set_host(Host),  da prob in locale
	set_that_port(Port),
	there,
	out(laila_query(Q,Agent, Goal, AbdIn)),
	println(['communication:(', Agent,Goal,AbdIn,')']), nl,
	com(Q,Agent,AbdIn,AbdOut).

com(Q,Agent,AbdIn,AbdOut):-
	iAm(WhoAmI),
	agent(WhoAmI,Host,Port),
	set_that_port(Port),
 	there,
	in(ris(Q,no)), % aspetto che l'agente remoto abbia finito di produrre risposte.
	all(ris(Q,Abd),L),
	println(L),nl,
	member(ris(Q,the(Abd)),L), % creo i punti di scelta per il backtracking
	println(['Risultato comunication prima del check di consistenza: (',Abd,')']), nl,
	consistency([WhoAmI,Agent], Abd, AbdIn, AbdOut),
	println(['Risultato comunication: (',AbdOut,')']), nl.
	
% competitive
laila(or(G1,G2), AbdIn, AbdOut):-
	println('laila: solving competitive'), nl,
	solveor(G1,G2,AbdIn,AbdOut).

solveor(G1,_,AbdIn,Abd):-
	laila(G1,AbdIn,Abd),
	println('Partecipante alla competizione left ha avuto successo'),
	println(Abd), nl.

solveor(_,G2,AbdIn,Abd):-
	laila(G2,AbdIn,Abd),
	println('Partecipante alla competizione right ha avuto successo'),
	println(Abd), nl.
	
% collaborative
laila(and(G1,G2), AbdIn, AbdOut):-
	println('laila: solving collaborative'), nl,
	laila(G1,AbdIn,Abd1),
	laila(G2,Abd1,Abd2),
	println(['Risultato collaborazione da fare ancora consistenza: (',Abd2,')']), nl,
	agent_bunch(G1,G2,Ag),
	iAm(WhoAmI),
	consistency([WhoAmI|Ag],Abd2,AbdIn,AbdOut),
	% verifico consistenza con le query
	query(G1,LG1),
	query(G2,LG2),
	union(LG1,LG2,LG),
	no_conflicts(LG,AbdOut),
	println(['Risultato collaborazione: (',AbdOut,')']), nl.

% consistency
consistency(Bunch, Delta, DeltaIn, DeltaOut):-
	println(['laila: consistency check: consistency(',Bunch,Delta,DeltaIn,DeltaOut,')']), nl,
	ask_consistency(Bunch, Delta, DeltaIn, DeltaOut).


% Lancio run_server
server:-
	iAm(WhoAmI), % my name:  nella kb dell'agente
	agent(WhoAmI, Host, Port), %  nella kb dell'agente
	set_this_port(Port),
	run_server.

agent_bunch(G1,G2,Ag):-
	g_bunch(G1,Ag1),
	g_bunch(G2,Ag2),
	union(Ag1,Ag2,Ag).

g_bunch(and(G1,G2),Ag):-
	g_bunch(G1,Ag1),
	g_bunch(G2,Ag2),
	union(Ag1,Ag2,Ag).
g_bunch(ask(A,_),[A]).
g_bunch(G,[]):-not(eq(G,and(_,_))),not(eq(G,ask(_,_))).

label_comm(Q):-
	iAm(WhoAmI),
	agent(WhoAmI, Host, Port),
	set_that_port(Port),
 	there,
	in(cont(N)),
	is(N1,'+'(N,1)),
	out(cont(N1)),
	eq(Q,[comm,WhoAmI,N1]).

% Questa clausola va messa per non aver problemi con le query di un'agente che non ha nemmeno
% una clausola del tipo abm(x,x).Infatti in questo caso STRANAMENTE si hanno messaggi di errore.
% Sembra che unifichi con la clausola della normal deduction, poi non trovi nessuna abm(x,x)
% dando un messaggio di errore. Il tutto per poi proseguire normalmente.
abm(no_fail,[no_fail]).

senza_label:-
in(laila_query(WhoAmI,LailaQuery,AbdIn)),
eq(Q,1), 
out(laila_query(Q,WhoAmI,LailaQuery,AbdIn)).

is_list([]).
is_list([H|T]).

query(and(G1,G2),LG):-
	query(G1,LG1),
	query(G2,LG2),
	union(LG1,LG2,LG).
query(ask(_,G),LG):-
	query(G,LG).
query(down(G),LG):-
	query(G,LG).
query(G,LG):-
	not(eq(G,and(_,_))),
	not(eq(G,ask(_,_))),
	not(eq(G,down(_))),
	if(
	   is_list(G),
	   eq(LG,G),
	   eq(LG,[G])
	  ).



