% ABM con collaborazione e competizione in parallelo (background).
% Prende la prima soluz della competizione che ha avuto successo. 
% Per avere un vero parallelismo devo mandare i thread in background con bg() e non posso usare new_engine().

% ABM: LAILA Metainterpreter for ALIAS AGENTS: si occupa della comunicazione tra agenti.

% Ogni ALIAS AGENTS ha una parte come questa che accede alla blackboard dell'IP e si occupa
% della comunicazione tra agenti tramite linguaggio lAILA:  l'ABM che si mette in ascolto
% con una in(laila_query(WhoAmI,LailaQuery,AbdIn)) di eventuali richieste di altri ALIAS AGENT
% nella forma out(laila_query(WhoAmI,LailaQuery,AbdIn)).
% ES: se l'agente 'pippo' sull'IP1 vuole chiedere all'agente 'pluto' sull'IP2 di risolvere
% un goal G 'pippo' dovr effettuare una laila query di tipo comunicativo ( 'pluto' > G ).
% Nella knowledge base di 'pippo' ci sar l'informazione sull' IP in cui si trova 'pluto'
% quindi  l'ABM di 'pippo' far una out() remota che verr ricevuta dall'ABM di 'pluto' del tipo
% out(laila_query('pluto', G, AbdIn)).

% Se si vogliono usare pi agenti su di uno stesso IP, dal momento che non posso far 
% partire pi run_server sulla stessa porta devo usare oltre agli indirizzi IP anche
% le potre. Quindi fare set_this_port() prima di run_server e per identificare un'agente
% usare agent(Agent, Host, Port).

% LAILA server: fetches laila queries from the blackboard
% queries are in the form of tuples: laila_query(<MyName>,Query,InputDelta)
% The agent has got one blackboard (1 agent per IP in the system); the
% agent is the blackboard server, it fetches queries by means of 'in'
% primitives and solves them, putting an answer on the blackboard.
% answers to queries are tuples of the form: answer(<MyName>,Query,InputDelta,OutputDelta).
% for instance: laila_query('torroni',s1,[]), answer('torroni',s1,[],[m1])
% I assume that OutputDelta is funcition of (agent,query,inputDelta).

start:- bg(run_laila_server).

run_laila_server:-
	bg(server), % my blackboard is your blackboard
	iAm(WhoAmI), % my name:  nella kb dell'agente
	agent(WhoAmI, Host, Port), %  nella kb dell'agente
	set_that_port(Port),
 	there, % in the blackboard I share
	repeat, % loop...
	in(laila_query(WhoAmI, LailaQuery, AbdIn)), % fetch (blocking)
	nl,
	println('Fetched a query for me. Issuing background execution...'), nl,
	bg(solve_laila_query(WhoAmI, LailaQuery, AbdIn)), % background execution (non-blocking)
	fail. % ...forever

solve_laila_query(WhoAmI, LailaQuery, AbdIn):-
	new_engine(laila(LailaQuery, AbdIn, AbdOut), AbdOut, Engine), % issue a thread (the thread then begins to solve the query)
	println('new_engine'), nl,
	ask_engine(Engine, Ans), % ask the thread for one answer   
	if(
         % se la risposta  del tipo the(X)
	   eq(Ans,the(X)),
         % allora...
	   eq(Answer,X),
         % altrimenti la risp  NO
	   eq(Answer,Ans)
	  ), 
	println('ask_engine:'),
	println(Answer), nl,
	out(answer(WhoAmI, LailaQuery, AbdIn, Answer)), % once you get it, put the answer on the blackboard
	
      % here the agent could be asked more solutions...
	
      stop_engine(Engine). % I'm done with that thread


% LAILA metainterpreter: queries are in the form: laila(LailaQuery,AbdIn,AbdOut)

% termination
laila([], AbdIn, AbdIn):-
	println('laila: terminated, returning '),
	println(AbdIn), nl.

% unroll goal
laila([A|B], AbdIn, AbdOut):-
	println('laila: unrolling goal'), nl,
	laila(A, AbdIn, AbdTemp),
	laila(B, AbdTemp, AbdOut).

% normal deduction
laila(Atom, AbdIn, AbdOut):-
	println('laila: performing normal deduction'), nl,
	abm(Atom, Body), % le clausole LAILA sono di questo tipo. 
	laila(Body, AbdIn, AbdOut).

% down-reflection
laila(down(Goal), AbdIn, AbdOut):-
	println('laila: solving down reflection...'), nl,
	solve_ARM_goal(Goal, AbdIn, AbdOut),
	% verifico consistenza con la query
	if(
	   is_list(Goal),
	   no_conflicts(Goal,AbdOut),
	   no_conflicts([Goal],AbdOut) 
	  ).

% communication
laila(ask(Agent, Goal), AbdIn, AbdOut):-
	println('laila: solving commuication...'), nl,
	agent(Agent, Host, Port), %  nella kb dell'agente
	% set_host(Host),  da prob in locale
	set_that_port(Port),
	there,
	out(laila_query(Agent, Goal, AbdIn)),
	println(['communication:(', Agent,Goal,AbdIn,')']), nl,
	in(answer(Agent, Goal, AbdIn, Abd)),
	iAm(WhoAmI),
	println(['Risultato comunication prima del check di consistenza: (',Abd,')']), nl,
	consistency([WhoAmI,Agent], Abd, AbdIn, AbdOut),
	println(['Risultato comunication: (',AbdOut,')']), nl.

% competitive
laila(or(G1,G2), AbdIn, AbdOut):-
	println('laila: solving competitive'), nl,
	bg(solve_formula(left,G1,_,AbdIn)),
	println('background or 1'), nl,
	bg(solve_formula(right,_,G2,AbdIn)),
	println('background or 2'), nl, 
	iAm(WhoAmI), % my name:  nella kb dell'agente
	agent(WhoAmI, Host, Port), %  nella kb dell'agente
	set_that_port(Port),
 	there, % in the blackboard I share
	in(fineor(LR,G1,G2,AbdIn,Abd,XY)),
	case_or(LR,XY,Abd,AbdOut).

% Il primo risultato arrivato  un successo.
case_or(LR,1,Abd,Abd):-
	bg(in(fineor(LRS,G1,G2,AbdIn,AbdS,XYS))), % tiro su in background il risultato che arriva secondo cos non resta sulla blackboard 
	println('Primo arrivato con successo'),
	println(LR),
	println('con risultato'),
	println(Abd),nl.

% Il primo risultato arrivato  un fallimento.
case_or(_,0,_,AbdOut):-
	in(fineor(LRS,G1,G2,AbdIn,AbdS,XYS)),
	case_or_2(LRS,XYS,AbdS,AbdOut).

case_or_2(LRS,1,AbdS,AbdS):-
	println('Primo arrivato con successo'),
	println(LRS),
	println('con risultato'),
	println(AbdS),nl.

case_or_2(_,0,_,_):-
	println('Nessuno dei 2 ha avuto successo'),nl,
	fail.

% collaborative
laila(and(G1,G2), AbdIn, AbdOut):-
	println('laila: solving collaborative'), nl,
	bg(solve_Sformula(left,G1,_,AbdIn)),
	println('background and 1'), nl,   
	bg(solve_Sformula(right,_,G2,AbdIn)),
	println('background and 2'), nl,
	iAm(WhoAmI), % my name:  nella kb dell'agente
	agent(WhoAmI, Host, Port), %  nella kb dell'agente
	set_that_port(Port),
 	there, % in the blackboard I share
	in(fineand1(G1,AbdIn,Abd1)),
	in(fineand2(G2,AbdIn,Abd2)),
	println(['Risultato delle 2 collab in background: (',Abd1,Abd2,')']), nl,   
	union(Abd1,Abd2,Abd),
	println(['Risultato collaborazione da fare ancora consistenza: (',Abd,')']), nl,
	agent_bunch(G1,G2,Ag),
	iAm(WhoAmI),
	consistency([WhoAmI|Ag],Abd,AbdIn,AbdOut),
	% verifico consistenza con le query
	query(G1,LG1),
	query(G2,LG2),
	union(LG1,LG2,LG),
	no_conflicts(LG,AbdOut),
	println(['Risultato collaborazione: (',AbdOut,')']), nl.

% consistency
consistency(Bunch, Delta, DeltaIn, DeltaOut):-
	println(['laila: consistency check: consistency(',Bunch,Delta,DeltaIn,DeltaOut,')']), nl,
	ask_consistency(Bunch, Delta, DeltaIn, DeltaOut).


% Lancio run_server
server:-
	iAm(WhoAmI), % my name:  nella kb dell'agente
	agent(WhoAmI, Host, Port), %  nella kb dell'agente
	set_this_port(Port),
	run_server.

agent_bunch(G1,G2,Ag):-
	g_bunch(G1,Ag1),
	g_bunch(G2,Ag2),
	union(Ag1,Ag2,Ag).

g_bunch(and(G1,G2),Ag):-
	g_bunch(G1,Ag1),
	g_bunch(G2,Ag2),
	union(Ag1,Ag2,Ag).
g_bunch(ask(A,_),[A]).
g_bunch(G,[]):-not(eq(G,and(_,_))),not(eq(G,ask(_,_))).

solve_Sformula(LR,G1,G2,AbdIn):-
	if(
	   eq(LR,left),
	   laila(G1,AbdIn,Abd1),
	   laila(G2,AbdIn,Abd2)
	  ),
	iAm(WhoAmI), % my name:  nella kb dell'agente
	agent(WhoAmI, Host, Port), %  nella kb dell'agente
	set_that_port(Port),
 	there, % in the blackboard I share
	% La out() che segue serve per 2 motivi:
	% 1)per passare Abd1 fuori dal thread.
	% 2)per sincronizzarmi tramite in() con il processo generatore.
	% Da notare G1 e AbdIn per far prender su alle in() le giuste out().
	if(
	   eq(LR,left),
	   out(fineand1(G1,AbdIn,Abd1)),
	   out(fineand2(G2,AbdIn,Abd2))
	  ).

solve_formula(LR,G1,G2,AbdIn):-
	if(
	   eq(LR,left),
	   if(
	      laila(G1,AbdIn,Abd),
	      eq(X,1),
	      eq(X,0)
	     ),
	   if(
	   laila(G2,AbdIn,Abd),
	   eq(Y,1),
	   eq(Y,0)
	     )
	  ),
	iAm(WhoAmI), % my name:  nella kb dell'agente
	agent(WhoAmI, Host, Port), %  nella kb dell'agente
	set_that_port(Port),
 	there, % in the blackboard I share
	% La out() che segue serve per 2 motivi:
	% 1)per passare Abd1 fuori dal thread.
	% 2)per sincronizzarmi tramite in() con il processo generatore.
	% Da notare G1 e AbdIn per far prender su alle in() le giuste out().
	if(
	   eq(LR,left),
	   out(fineor(left,G1,G2,AbdIn,Abd,X)),
	   out(fineor(right,G1,G2,AbdIn,Abd,Y))
	  ).

is_list([]).
is_list([H|T]).

query(and(G1,G2),LG):-
	query(G1,LG1),
	query(G2,LG2),
	union(LG1,LG2,LG).
query(ask(_,G),LG):-
	query(G,LG).
query(down(G),LG):-
	query(G,LG).
query(G,LG):-
	not(eq(G,and(_,_))),
	not(eq(G,ask(_,_))),
	not(eq(G,down(_))),
	if(
	   is_list(G),
	   eq(LG,G),
	   eq(LG,[G])
	  ).



