% ARM: Metainterprete per la realizzazione della derivazione abduttiva degli ALIAS AGENT
% realizzato usando l'algoritmo di Kakas-Mancarella esteso al multi-agent.
% La risoluzione abduttiva  locale nel senso che l'agente utilizza solo la propria
% knowledge base.A livello pi alto (nell' ABM) ci si occuper di ottenere o meno
% una eventuale consistenza globale a posteriori.Nella realizzazione precedente di ALIAS
% avevo anche l una abduzione locale ma la consistenza era controllata man mano che
% creavo un Delta comune. Questo aveva tuttavia lo svantaggio di non permettere
% competizione. 


solve_ARM_goal(Goal, AbdIn, AbdOut):-
	solve_local_abd(Goal, AbdIn, AbdOut).


% Algoritmo di Kakas-Mancarella esteso al multi-agent. Da SAC'98:
% E' diviso in 2 parti: - 1) Abductive derivation
%                       - 2) Consistency derivation
%    
%
% 1) Abductive derivation 
%
% An abductive derivation  from (G_1 Delta_1) to (G_n Delta_n)
% for an agent A0 (with arguing agents A_1, ..., A_m) with
% knowledge base given by the abductive logic program
% <P,Ab,IC> via a selection rule R is a sequence
% (G_1 Delta_1),  (G_2 Delta_2), ..., (G_n Delta_n)
% such that each Gi has the form <- L_1, ..., L_k,
% R(G_i)=L_j and (G_{i+1} Delta_{i+1}) is obtained according to
% one of the following rules:
% 
% [(A1)]	If L_j is not abducible, then G_{i+1}=C and Delta_{i+1} =
% 	Delta_i where C is the resolvent of some clause in P
% 	with G_i on the selected literal L_j;
% 
% [(A2)] 	If L_j is abducible and L_j in Delta_i then 
% 	G_{i+1} = <- L_1, ..., L_{j-1}, L_{j+1}, ..., L_k and
% 	Delta_{i+1} = Delta_i;
% 
% [(A3)]	If L_j is abducible, L_j not in Delta_i and 
% 	not(L_j) not in Delta_i and for each arguing agent A_k
% 	(k=1, ..,m) there exists a consistency derivation from 
% 	(L_j Delta_i U L_j) to ({} Delta'_{k}) and  the union set
% 	U_{k=1, ..., n} Delta'_{k} is consistent (i.e., L and not(L)
% 	do not belong to the union set for any literal L), then
% 	G_{i+1} = <- L_1, ..., L_{j-1}, L_{j+1}, ..., L_k and
% 	Delta_{i+1} = U_{k=1, ..., n} Delta'_{k}.
%
%
% 2) Consistency derivation 
%
% Given an agent A_0 (with arguing agents A_1, ..., A_m) whose
% knowledge base is given by the abductive logic program
% <P,Ab,IC>, a consistency derivation  for an abducible
% literal alpha from (alpha Delta_1) to (F_n Delta_n) is a sequence
% (alpha Delta_1), (F_1 Delta_1), (F_2 Delta_2), ..., (F_n Delta_n)
% where :
% 
% [(C_i)] F_1 is the union of all goals of the form <- L_1, ..., L_n
% 	obtained by resolving the abducible alpha with the denials in IC 
% 	with no such goal been empty, <-;
% 
% [(C_ii)] for each i > 1,  F_i has the form {<- L_1, ..., L_k} U F_i'
% 	and for some j=1, ..., k (F_{i+1} Delta_{i+1}) is obtained according to one
% 	of the following rules:
% 	[(C1)] If L_j is not abducible, then 
% 		F_{i+1} = C' U F_i' where C' is the set of all resolvents of 
% 		clauses in P with <- L_1, ..., L_k on the literal
% 		L_j and <- not in C', and Delta_{i+1} = Delta_i U not(L_j);
% 	[(C2)] If L_j is abducible, L_j in Delta_i and k>1, then 
%         	F_{i+1}={<- L_1, ..., L_{j-1}, L_{j+1}, ..., L_k} U F_i' and
% 		Delta_{i+1} = Delta_i;
% 	[(C3)] If L_j is abducible, not(L_j) in Delta_i then 
% 		F_{i+1} = F_i' and Delta_{i+1} = Delta_{i};
% 	[(C4)] If L_j is abducible, L_j not in Delta_i and not(L_j) not in Delta_i,
% 		and there exists a (local) abductive derivation from (<- not(L_j) Delta_ i)
% 		to (<- Delta') then F_{i+1} = F_i' and \Delta_{i+1} = \Delta'.



% Realizzazione algoritmo:



% 1) Abductive derivation 

%Di seguito la versione per prolog standard in forma di commento poi quella per Jinni
%
% 	solve_local_abd(Goal,AbdIn,AbdOut).
%
% /* [(termination)] */
% solve_local_abd([],Abd,Abd):-!.
%
% /* [(unroll goal)] */
% solve_local_abd([A|B],AbdIn,AbdOut):-
% 	!, solve_local_abd(A,AbdIn,Abd1),
%	solve_local_abd(B,Abd1,AbdOut).
%
% /* [(A1)] A is not abducible => normal deductive resolution (solve Body) */
% solve_local_abd(A,AbdIn,AbdOut):-
% 	not(abducible(A)),
% 	rule(A,Body), % nella versione precedente di ALIAS le regole si chiamavano 'rule'
% 	solve_local_abd(Body,AbdIn,AbdOut).
%
% /* if bumped into an abducible: */
%
% /* [(A2)] A has already been abduced */
% solve_local_abd(A,Abd,Abd):- 	% /* Nota: A pu essere positivo o negativo */
% 	abducible(A),
% 	member(A,Abd),!.
%
% /* [(A3)] A is abducible but neither A nor not(A) have already been abduced */
% solve_local_abd(A,AbdIn,AbdOut):-
%	abducible(A),
% 	not(abduced(A,AbdIn)),
% 	% qua l'unica differenza col caso distribuito: non viene eseguito un consistency check
% 	% con gli altri agenti; il consistency check  eseguito localmente.
% 	solve_local_con(A,[A|AbdIn],AbdOut).



solve_local_abd(A, AbdIn, AbdOut):-
	println(['solve_local_abd(', A, AbdIn, AbdOut, ')']), nl,
	if(
   	   % se A=[]
         eq(A, []),
         % allora termino
         eq(AbdIn, AbdOut), 
         % altrimenti...
         if(
            % se devo fare unroll del goal 
            eq(A, [A1|A2]),
            % allora...
	      and(
                solve_local_abd(A1, AbdIn, Abd1), 
	          solve_local_abd(A2, Abd1, AbdOut) 
               ),
            % altrimenti..... 
            if(
               % se A non  abducibile
               not(abducible(A)),
               % allora ho la risoluzione normale del Body
	         and(
                   arm(A, Body),
		       solve_local_abd(Body, AbdIn, AbdOut)
                  ),
               % altrimenti A  abducibile quindi...
	         if(
                  % se A  gi stato abdotto 
                  abduced(A, AbdIn),
                  % allora... 
		      eq(AbdIn, AbdOut),
                  % altrimenti se ne A ne not(A) sono gi stati abdotti...
		      and(
           	          not(abduced_(A, AbdIn)), % otherwise: A is abducible but neither A nor not(A) have already been abduced
		          solve_local_con(A, [A|AbdIn], AbdOut)
                     )
	           )
	        )
           )
      ).



% 2) Consistency derivation */

% /* solve consistency; finds out all the ICs containing Goal - which HAS TO BE a literal, 
% either negated or not and makes a list out of them, List_of_ic. If this list turns out to
% be empty then there's no IC to satisfy and it's finished (i.e., the Goal can be abduced), 
% otherwise it appends not(Goal) to it; after producing this way All_ic, it returns control 
% to solve_con2. solve_con2 selects an L_j (i.e., implements a selection rule is order to
% allow backtracking when ICs have got more than 2 elements), and calls solve_con3. */


solve_local_con(Goal, AbdIn, AbdOut):-
	println(['solve_local_con(', Goal, AbdIn, AbdOut, ')']), nl,
	if(
         % se esistono dei vincoli per il goal
	   and(
	       findall(Ic, get_ic(Goal, Ic), List_of_ic),
		 not(eq(List_of_ic, [])) % List_of_ic \== []
		), 
	   % allora... 
	   and(
		 add_ic(Goal, List_of_ic, All_ic),
		 solve_local_con2(All_ic, AbdIn, AbdOut)
		), 
	   % altrimenti...
	   eq(AbdIn, AbdOut)
	  ).


% Versione prolog standard:
% solve_con2: termination (no more ICs to satisfy) 
% solve_local_con2([],Abd,Abd):-!.
%
% solve_local_con2(ICs,AbdIn,AbdOut):-
% 	select_Lj(ICs,NewICs), % here it should backtrack (which didn't happen previously)
% 	solve_local_con3(NewICs,AbdIn,AbdOut).

solve_local_con2(ICs, AbdIn, AbdOut):-
      if(
         % se non ho pi vincoli
         eq(ICs, []),
         % allora...
	   eq(AbdIn, AbdOut),
         % altrimenti li risolvo...
	   and(
             select_Lj(ICs, NewICs),
             solve_local_con3(NewICs, AbdIn, AbdOut)
            )
	  ).

% solve_con3: termination (no more ICs to satisfy)
solve_local_con3([], Abd, Abd).                           

% /* [(C1)] L1 is not abducible, so try to use normal resolution */
solve_local_con3([[L1|Rest_l]|Rest_ic], AbdIn, AbdOut):-
	not(abducible(L1)),
      if(
	   % se L1  la testa di qualche clausola di P
	   and( 
	       findall(Body, arm(L1,Body), List_of_Bodies),
	       not(eq(List_of_Bodies, [])) % List_of_Bodies \== [] 
		),
         % allora vado con la normale risoluzione abduttiva
	   and( 
	       and(
                 append_all(List_of_Bodies, Rest_l, New_list),
	           append(New_list, Rest_ic, New_ic) ),
		     solve_local_con2(New_ic, [not(L1)|AbdIn], AbdOut)
		    ),
         % altrimenti procedo controllando gli altri vincoli (questo vincolo  gi fallito)
	   solve_local_con2(Rest_ic, [not(L1)|AbdIn], AbdOut)
	  ).

% /* [(C2)] L1 has already been abduced */
solve_local_con3([[L1|Rest_l]|Rest_ic], AbdIn, AbdOut):-
	abduced(L1, AbdIn),
	solve_local_con3([Rest_l|Rest_ic], AbdIn, AbdOut).

% /* [(C3)] not(L1) has been abduced */
solve_local_con3([[L1|_]|Rest_ic], AbdIn, AbdOut):-
	abducible(L1),
	negate(L1, L1Neg),
	abduced(L1Neg, AbdIn),
	solve_local_con2(Rest_ic, AbdIn, AbdOut).

% /* [(C4)] L1 is abducible but neither L1 nor not(L1) have already been abduced */
solve_local_con3([[L1|_]|Rest_ic], AbdIn, AbdOut):-
	abducible(L1),
	not(abduced_(L1, AbdIn)),
	negate(L1, L1Neg),
	solve_local_abd(L1Neg, AbdIn, AbdOut1),
	solve_local_con2(Rest_ic, AbdOut1, AbdOut).



% Funzioni varie:


% Mi dice se A  abducibile
abducible(A):-
      if(
         eq(A, not(NotA)),
	   true,
	   and(abducibles(AbdSet), member(A, AbdSet))
	  ).

% Mi dice se A  gi stato abdotto
abduced(A, Delta):-
	member(A, Delta).

% Mi dice se A o NotA sono gi stati abdotti
abduced_(A, Delta):- 
	negate(A, ANeg),
	or(
	   abduced(A, Delta),
         abduced(ANeg, Delta)
        ).

% get_ic(Goal,Ic) produce in Ic una lista costituita dal body di un IC contenente Goal,
% privato dello stesso Goal 
get_ic(Goal, Ic):-
	ic(X),
	member(Goal, X),
	delete(X, Goal, Ic).

% add_ic(Goal, ListIn, ListOut) produce una lista ListOut costituita da Goal negato, in
% testa, e da ListIn. ListIn  una lista di liste: [[abd_1, abd_2, ...], ...,[abd_j, abd_k, 
% ...]] con abd_X della forma L_i oppure not(L_i)
% Vers prolog standard:
% add_ic(not(Goal),List,[[Goal]|List]):-!.
% add_ic(Goal,List,[[not(Goal)]|List]).
add_ic(Goal, List, [[NewGoal]|List]):-
	if(
         eq(Goal, not(NotGoal)),
	   eq(NewGoal, NotGoal),
	   eq(NewGoal, not(Goal))
	  ).

% delete(ListIn, Elem, ListOut) restituisce in ListOut la lista di input ListIn privata 
% di tutte le istanze dell'elemento Elem
% Vers prolog standard:
% delete([],_,[]):-!.
% delete([H|Rest],H,Rest1):-!,delete(Rest,H,Rest1).
% delete([H|Rest],Goal,[H|Rest1]):-!,delete(Rest,Goal,Rest1).
delete([], _, []).
delete([H|Rest], Goal, Rest1):-
	delete(Rest, Goal, Rest2),
	if(
         eq(Goal, H),
	   eq(Rest1, Rest2),
	   eq(Rest1, [H|Rest2])
        ).

% select_Lj(ICs,NewICs) imlements a selection rule for L_j. Given a set ICs of n sets, 
% being m the cardinality of ICs[1], it produces at each backtracking one out of m different
% permutations obtained by a shift and rotate of ICs[1] itself. It doesn't affect the other
% elements (lists) of ICs. It fails after m times.
select_Lj([Head|ICs], [NewHead|ICs]):-
	cardinality(Head, N),
	rotate(Head, N, NewHead).

% cardinality(Set,N) returns N as the cardinality of Set.
cardinality([], 0).
cardinality([Head|Tail], N):-
	cardinality(Tail, N1),
	is(N, '+'(N1,1)).

% rotate(List,N,NewList) rotates List N times, afterwards it fails. The first time it is
% invoked, Newlist is but List; the other times NewList is obtained by appending the head of
% List on its tail.
rotate(X, N, X):-	less(0, N). % N > 0.
rotate([Head|Tail],N,NewList):-
	less(0, N), % N > 0,
	is(N1, '-'(N,1)), % N1 is N - 1,
	append(Tail, [Head], Rotated),
	rotate(Rotated, N1, NewList).

% negate(Abd,AbdNeg) produce in AbdNeg la negazione di Abd.
% Vers prolog standard:
% negate(not(L1),L1):-!.
% negate(L1,not(L1)).
negate(L1, L2):-
	if(
         eq(L1, not(NotL1)),
	   eq(L2, NotL1),
	   eq(L2, not(L1))
	  ).

% append_all(ListIn, Elem, ListOut) appende Elem in testa a ogni elemento
% di ListIn (che  una lista di liste), producendo cos a sua volta una lista 
% di liste: ListOut.
append_all([], _, []).
append_all([Body|Rest_b], Rest_l, [New_clause|Rest_c]):-
     append(Body, Rest_l, New_clause),
     append_all(Rest_b, Rest_l, Rest_c).

% append(X,Y,Z) attacca X dopo Y producendo Z
append([], X, X).
append([A|X], Y, [A|Z]):-
	append(X, Y, Z).

% Jinni ha gi una member. E comunque fatta con if() non ha punti di scelta.
% member(List, Elem) torna true se e solo se Elem fa parte di List
% Vers prolog standard:
% member(X,[X|_]):-!.
% member(X,[_|Z]):-member(X,Z).
%member(X, [H|T]):-
%	if(
%         eq(X, H),
%	   true,
%	   member(X, T)
%	  ).

% Ha lo stesso scopo di abm(no_fail,[no_fail]) in ABM-interpreterMulti.pro
ic([buddy_ic1, buddy_ic2]).                                  

